--[[
	Whack-That-Toe is a realtime competitive version of Tic-Tac-Toe.
	By Philipp Überbacher aka. murks.
]]--

--[[
grid = {
  {{1}, {2}, {3}},
  {{4}, {5}, {6}},
  {{7}, {8}, {9}},
}
]]--

-- can't set this too early
love.graphics.setDefaultFilter('nearest', 'nearest')

require 'lib/global_lock'
GLOBAL_lock(_G)
local Class = require 'lib/middleclass'
local Timer = require 'lib/timer'
local State = require 'lib/gamestate'
local Suit = require 'lib/suit'


-- game state tables
local menu = {}
local game = {}
local win = {}

local triples = {
	{1,2,3},
	{4,5,6},
	{7,8,9},
	{1,4,7},
	{2,5,8},
	{3,6,9},
	{1,5,9},
	{3,5,7},
}

local wasd_codes = {
	q = 1, w = 2,	e = 3,
	a = 4, s = 5,	d = 6,
	z = 7, x = 8,	c = 9,
}

local TOE = love.math.random(1, 9)
local WINNER = {}
local FONT = {}
local KEYLOCK = false
local SFX = {}

local cells = {}
local Cell = Class('Cell')
Cell.static.toe = '@'
Cell.static.x = 'x'
Cell.static.o = 'o'
Cell.static.empty = ''

local function match(t)
	-- empty cells should never match
	if cells[t[1]].symbol == Cell.static.empty then return false end
	-- do the three cells match?
	if cells[t[1]].symbol == cells[t[2]].symbol
	and cells[t[2]].symbol == cells[t[3]].symbol
	then return {cells[t[1]], cells[t[2]], cells[t[3]]}
	else return false end
end

local function keypad(key)
	if key:match('^kp%d') then
		local k =  tonumber(key:sub(3))
		if k > 0 and k < 10 then
			-- weird hack to invert keypad
			if k > 6 then
				k = k-6
			elseif k < 4 then
				k = k+6
			else
				-- center row can stay as it is
			end
			cells[k]:mark(Cell.static.x)
		end
	end
end

local function wasd(scancode)
	if scancode:len() == 1 and scancode:match('^%a') then
		local index = wasd_codes[scancode]
		if index then cells[index]:mark(Cell.static.o) end
	end
end

local function hit_cells(x, y)
	for _,c in pairs(cells) do
		if c:hit(x, y) then
			c:mark(Cell.static.x)
		end
	end
end


Cell.static.check = function()
	-- primitive way to check but no one needs smarts on this scale
	local winner_found = false
	for _,t in pairs(triples) do
		local triple = match(t)
		if triple then
			WINNER = triple
			KEYLOCK = true
			SFX.win:play()
			Timer.after(1, function()
					KEYLOCK = false
					winner_found = true
					State.switch(win)
				end)
		end
	end
	if not winner_found and Cell.static.is_full() then
		State.switch(win)
	end
end

Cell.static.is_full = function()
	local full = true
	for _,v in pairs(cells) do
		if v.symbol == Cell.static.empty then
			full = false
		end
	end
	return full
end

Cell.static.reset = function()
	for _,v in pairs(cells) do v.symbol = Cell.static.empty end
	WINNER = {}
end

Cell.static.get_empty = function()
	local empty_cells = {
		n = 0
	}
	for k,v in ipairs(cells) do
		if v.symbol == Cell.static.empty then
			empty_cells[k] = true
			empty_cells.n = empty_cells.n + 1
		end
	end
	return empty_cells
end

Cell.static.move_toe = function()
	local empty = Cell.static.get_empty()
	if empty.n < 1 then
		return -- important to bail here
	end
	-- remove toe
	if cells[TOE].symbol == Cell.static.toe then
		cells[TOE].symbol = Cell.static.empty
	end
	-- determine next toe position
	local rand
	repeat
		rand = love.math.random(1,9)
	until empty[rand]
	-- set new toe
	TOE = rand
	cells[rand].symbol = Cell.static.toe
end

function Cell:initialize(index)
	self.size = love.graphics.getWidth()/3 -- window width is 360 px
	self.index = index
	self.x = ((self.index-1) % 3) * self.size -- not sure this works in general
	self.y = (math.ceil(self.index/3)-1) * self.size -- does work for 3x3 though :}
	self.symbol = Cell.static.empty
end

function Cell:mark(symbol)
	if self.symbol == Cell.static.toe then
		self.symbol = symbol
		Cell.static.check()
		SFX.hit:setPitch(love.math.random(80, 120)/100)
		SFX.hit:play()
	else
		SFX.miss:setPitch(love.math.random(80, 120)/100)
		SFX.miss:play()
	end
end

function Cell:draw()
	love.graphics.rectangle('line', self.x, self.y, self.size, self.size)
	if self.symbol == Cell.static.toe then
		love.graphics.setColor(.2, .6, .73)
		love.graphics.print(self.symbol, self.x + 24, self.y + 11)
		love.graphics.setColor(1, 1, 1)
	else
		love.graphics.print(self.symbol, self.x + 24, self.y - 3)
	end
end

function Cell:hit(x, y)
	if x > self.x
	and x < self.x + self.size
	and y > self.y
	and y < self.y + self.size
	then return true
	else return false end
end


function love.load()
	if arg[#arg] == '-debug' then require('mobdebug').start() end

	FONT.large = love.graphics.newFont('font/monogram.ttf', 126, 'mono')
	FONT.small = love.graphics.newFont('font/monogram.ttf', 27, 'mono')

	SFX.select = love.audio.newSource('sfx/select.wav', 'static')
	SFX.select:setVolume(.3)
	SFX.hit = love.audio.newSource('sfx/hit.wav', 'static')
	SFX.hit:setVolume(.5)
	SFX.miss = love.audio.newSource('sfx/miss.wav', 'static')
	SFX.miss:setVolume(.2)
	SFX.win = love.audio.newSource('sfx/win.wav', 'static')
	SFX.win:setVolume(.5)

	Suit.theme.color = {
		normal  = {bg = { .25, .25, .25 }, fg = { 1, 1, 1 }},
		hovered = {bg = { .2, .6, .73 },   fg = { 1, 1, 1 }},
		active  = {bg = { 1, .6,  0 },     fg = { .87, .87, .87 }}
	}

	for i=1,9 do
		cells[i] = Cell:new(i)
	end

	Timer.every(1,
		function()
			Cell.static.move_toe()
		end)

	State.registerEvents()
	State.switch(menu)
end

function love.update(_)
end

function love.draw()
end

function love.keypressed(key)
	if key == 'escape' then	love.event.quit()	end
end


function menu:enter()
	love.graphics.setFont(FONT.small)
	menu.suit = Suit.new()
end

function menu:update(_)
	menu.suit:Label("Whack-That-Toe", 30,60, 300,20)

	love.graphics.setFont(FONT.large)
	menu.suit:Label("@", 138,145, 100,100)

	love.graphics.setFont(FONT.small)
	-- Put a button on the screen. If hit, show a message.
	if menu.suit:Button("Whack", {id=1}, 80,270, 200,60).hit then
		SFX.select:play()
		State.switch(game)
	end
end

function menu:draw()
	menu.suit:draw()
end

function menu:keypressed(key)
	if key == 'return' or key == 'kpenter' then
		SFX.select:play()
		State.switch(game)
	end
end


function game:enter()
	love.graphics.setFont(FONT.large)
	Cell.static.reset()
end

function game:update(dt)
	Timer.update(dt)
end

function game:draw()
	love.graphics.setColor(1, 1, 1)
	for _,v in pairs(cells) do v:draw() end
	if #WINNER > 0 then
		love.graphics.setColor(0, 1, 0)
		for _,c in pairs(WINNER) do
			c:draw()
		end
	end
end

function game:keypressed(key, scancode)
	if not KEYLOCK then
		keypad(key)
		wasd(scancode)
	end
end

function game:mousepressed(x, y)
	if not KEYLOCK then
		-- takes care of both mouse and touch input, neat!
		hit_cells(x, y)
	end
end


function win:enter()
	love.graphics.setFont(FONT.small)
	win.suit = Suit.new()
end

function win:update(_)
	win.suit:Label("Whack-That-Toe", 30,60, 300,20)

	if WINNER[1] then
		love.graphics.setFont(FONT.large)
		win.suit:Label(WINNER[1].symbol, 140,120, 100,100)
		love.graphics.setFont(FONT.small)
		win.suit:Label("WON!", 140,200, 100,80)
	else
		win.suit:Label("DRAW", 130,120, 100,100)
	end

	if win.suit:Button("Whack", {id=2}, 80,270, 200,60).hit then
		SFX.select:play()
		State.switch(game)
	end
end

function win:draw()
	win.suit:draw()
end

function win:keypressed(key)
	if key == 'return' or key == 'kpenter' then
		SFX.select:play()
		State.switch(game)
	end
end
